from fxq.commons._reflection_utils import _is_list_with_generic, _get_generic, _deserialize_json_list_to, \
    _deserialize_json_object_to, _is_list
from fxq.commons._requests_adapter import _do_get_json


def get(url, params=None, return_type=None, callback=None):
    """
    Sends a get request to the provided URL and deserializes the response.

    Lists are handled natively and types/callbacks are used within list elements.

    If return type is specified the json is deserialized using reflection.

    If callback is specified the json is passed to the callback function and its response is returned.

    DO NOT SPECIFY return_type AND callback AT THE SAME TIME

    :param url: Address to send get request to
    :param params: Parameters to pass in the request
    :param return_type: Class to deserialize the json into
    :param callback: Callback function which is invoked to deserialize the Json
    :return: Returns the result of mapping the Json to Type, or Callback response
    """
    if return_type and callback:
        raise Exception("Parameters not supported, use callback or returntype")

    if return_type:
        return _get_with_type(url, params, return_type)

    if callback:
        return _get_with_callback(url, params, callback)


def _get_with_type(url, params, return_type):
    json = _do_get_json(url, params)
    if _is_list_with_generic(return_type):
        return _deserialize_json_list_to(json, _get_generic(return_type))
    else:
        return _deserialize_json_object_to(json, return_type)


def _get_with_callback(url, params, callback):
    json = _do_get_json(url, params)
    if _is_list(json):
        return [callback(e) for e in json]
    else:
        return callback(json)



def _is_primitive(obj):
    return obj.__class__ in [str, float, int, bool]


def _is_list(obj):
    return obj.__class__ in [list]


def _is_json_object(obj):
    return obj.__class__ in [dict]


def _is_list_with_generic(type):
    return hasattr(type, '_name') and type._name == "List"


def _get_generic(signature):
    return signature.__args__[0]


def _deserialize_json_list_to(json_list, type):
    return [_marshal_dictionary_to_object(e, type) for e in json_list]


def _deserialize_json_object_to(json_obj, type):
    return _marshal_dictionary_to_object(json_obj, type)


def _marshal_dictionary_to_object(dictionary, object_type):
    resp_obj = object_type()
    for k, v in dictionary.items():
        if hasattr(resp_obj, k):
            if _is_primitive(v):
                setattr(resp_obj, k, v)
            elif _is_json_object(v):
                setattr(resp_obj, k, _marshal_dictionary_to_object(v, object_type.types[k]))
            elif _is_list(v):
                setattr(resp_obj, k, [_marshal_dictionary_to_object(e, _get_generic(object_type.types[k])) for e in v])
    return resp_obj



from stringcase import snakecase as beancase

from fxq.core.context.ApplicationContext import application_context


def Component(__component_class=None, name=None):
    '''
    Class Decorator to indicate that a bean of this class should be instantiated.
    Beans are created using the Class name by default and would be singletons.
    :param name: Optional Argument to define the explicit name of the bean
    :return: Returns the class after a bean has been instantiated and stored
    '''
    if __component_class:
        bean_id = beancase(__component_class.__name__)
        application_context.register_bean(bean_id, __component_class())
        return __component_class
    else:
        def wrapper(__component_class):
            application_context.register_bean(name if name else __component_class.__name__, __component_class())
            return __component_class

        return wrapper



import logging


class BeanFactory:
    LOGGER = logging.getLogger("BeanFactory")

    def __init__(self):
        self._beans = {}
        BeanFactory.LOGGER.info("Bean Factory Initialized")

    def register_bean(self, name, bean):
        BeanFactory.LOGGER.debug("Registering Bean for type %s with name %s" % (bean.__class__.__name__, name))
        self._beans[name] = bean

    def get_bean(self, name: str = None, required_type: type = None):
        if name and required_type is None:
            assert isinstance(name, str), "Expected bean name to be of type str, got \"%s\"" % name.__class__
            bean = self._get_bean_by_name(name)
            BeanFactory.LOGGER.debug(
                "Found instance of type %s for requested bean %s" % (bean.__class__.__name__, name))
            return bean
        elif required_type and name is None:
            return self._get_bean_by_type(required_type)
        elif name and required_type:
            return self._get_bean_by_name_and_validate_type(name, required_type)
        else:
            raise BeanFactoryException("get_bean requires at least name or required type")

    def _get_bean_by_name(self, name):
        try:
            return self._beans[name]
        except KeyError as e:
            BeanFactory.LOGGER.error("Failed to find Bean %s" % name, e)
            raise BeanFactoryException("BeanFactory contains no bean of name:%s" % name)

    def _get_bean_by_type(self, type):
        candidate_beans = []
        for bean_name, bean_instance in self._beans.items():
            if isinstance(bean_instance, type):
                BeanFactory.LOGGER.debug("Matched %s to be of class type %s" % (bean_name, type.__name__))
                candidate_beans.append(bean_instance)

        if len(candidate_beans) == 1:
            return candidate_beans[0]
        elif len(candidate_beans) > 1:
            raise BeanFactoryException("More than one bean for type %s was found" % type.__name__)
        elif len(candidate_beans) < 1:
            raise BeanFactoryException("No bean for type %s was found" % type.__name__)

    def _get_bean_by_name_and_validate_type(self, name, type):
        try:
            instance = self._beans[name]
            if isinstance(instance, type):
                BeanFactory.LOGGER.debug("Validated Bean %s to be of type %s" % (name, type.__name__))
                return self._beans[name]
            else:
                raise BeanFactoryException(
                    "Bean %s is of type %s expected %s" % (name, instance.__class__.__name__, type.__name__))
        except KeyError as e:
            BeanFactory.LOGGER.error("Failed to find Bean %s" % name)
            raise BeanFactoryException("Injection for Bean %s failed" % name)


class BeanFactoryException(Exception):
    pass
